
#include "common.hxx"
#include "basic.hxx"
#include "Config.hxx"
#include "Address.hxx"
#include "Input.hxx"
#include "Output.hxx"
#include "LongUDPSocket.hxx"
#include "ObjectPool.hxx"
#include "objects.hxx"
using namespace Rescue;

int main2(int argc, char** argv);
void commands(S32& time, Input& input, Output& output, LongUDPSocket& socket, Address& to, ObjectPool& pool, Config& config);
void commands(S32& time, Header type, Input& input, ObjectPool& pool, Config& config);
void load(S32& time, MovingObject* agent, Input& input, ObjectPool& pool);
void unload(S32& time, MovingObject* agent, Input& input, ObjectPool& pool);
void rescue(S32& time, MovingObject* agent, Input& input, ObjectPool& pool);
void stretch(S32& time, MovingObject* agent, Input& input, ObjectPool& pool, Config& config);
void move(S32& time, MovingObject* agent, Input& input, ObjectPool& pool, Config& config);
void clear(S32& time, MovingObject* agent, Input& input, ObjectPool& pool, Config& config);
void step(S32& time, ObjectPool& pool, Config& config);

int main(int argc, char** argv)
{
	try {
		return main2(argc, argv);
	} catch(std::exception& e) {
		fprintf(stderr, "%s\n", (const char*)e.what());
	} catch(...) {
		fprintf(stderr, "error...\n");
	}
	return 1;
}

int main2(int argc, char** argv)
{
	const char* host = "localhost";
	Config config;

	// 
	int i;
	int count = 0;
	for(i=1; i<argc; i++) {
		if(argv[i][0] == '-') {
			if(strcmp(argv[i], "-file") == 0) {
				if(++i < argc) {
					config.readConfigFile(argv[i]);
				} else {
					fprintf(stderr, "error: '-file'\n");
				}
			} else {
				if(++i < argc) {
					if(!config.setValue(argv[i-1]+1, argv[i]))
						fprintf(stderr, "error: '%s'\n", argv[i]);
				} else {
					fprintf(stderr, "error: '%s'\n", argv[i-1]);
				}
			}
		} else {
			switch(count++) {
			case 0:
				host = argv[i];
				break;
			default:
				fprintf(stderr, "error: '%s'\n", argv[i]);
				break;
			}
		}
	}

	// \Pbg쐬
	LongUDPSocket socket(config.textlogname_miscsimulator());
	Input input;
	Output output;
	output.setWait(config.send_udp_wait());
	Address to(host, config.port());	//M

	// ڑbZ[W𑗐M
	printf("connecting...\n");
	output.clear();								//o̓obt@ɂ
	output.put(SK_CONNECT);						//o̓obt@ SK_CONNECT (4oCg)o͂
	Output::Cursor base = output.put(~(S32)0);	//body ̃TCYƂĂƂ肠_~[̒lo͂ĂāAbase ɂ̈ʒuLB
	output.put(0);								//o̓obt@ 0 o͂
	output.setSize(base);						//base ŋLʒuɁAbase 炱܂ł put oCgݒ肷B
	output.put(HEADER_NULL);
	socket.send(output, to, -1);					//output  put oCg to ֑MB

	// bZ[W[v
	S32 time = 0;
	ObjectPool pool;
	for(;;) {
		try {
			socket.receive(input);	// TODO: G[OŕԂ	//f[^MAinput ֒~B
			for(;;) {
				Header header = input.get();	//input (4oCg)f[^oAheader ֑
				if(header == HEADER_NULL)
					break;

				S32 size = input.get();
				Input::Cursor start = input.cursor();	//ݓǂݏoĂʒuL

				switch(header) {
				default:
					printf("Unknown header.\n");
					break;
				case KS_CONNECT_ERROR:
					printf("Connection failed.\n");
					return 1;
				case KS_CONNECT_OK:
					printf("Connect Ok.\n");
					// ack Ԃ
					output.clear();
					output.put(SK_ACKNOWLEDGE);
					output.put(~(S32)0);
					output.put(HEADER_NULL);
					socket.send(output, to, -1);
					
					to = socket.addressRecievedFrom();	//ȍ~̑MύX
					pool.restructure(0, input);			//[hfB
					break;
				case KS_COMMANDS:
					commands(time, input, output, socket, to, pool, config);
					printf("time = %ld.\n", (long)time);
					break;
				case KS_UPDATE:
					time = input.get();
					pool.restructure(time, input);
					step(++time, pool, config);
					break;
				}

				input.setCursor(start);
				input.skip(size);						//size oCgXLbv
			}
		} catch(std::exception& e) {
			fprintf(stderr, "%s\n", (const char*)e.what());
		} catch(...) {
			fprintf(stderr, "error...\n");
		}
	}
	
	return 0;
}

void commands(S32& time, Input& input, Output& output, LongUDPSocket& socket, Address& to, ObjectPool& pool, Config& config)
{
	// KS_COMMANDS M
	time = input.get();
	Header type;
	while(type = input.get(), type != HEADER_NULL) {
		S32 size = input.get();
		Input::Cursor start = input.cursor();
		commands(time, type, input, pool, config);
		input.setCursor(start);
		input.skip(size);
	}

	// V~[Vʂ𑗐M
	output.clear();
	output.put(SK_UPDATE);
	Output::Cursor base = output.put(~(S32)0);
	pool.output(output, time);
	output.setSize(base);
	output.put(HEADER_NULL);
	socket.send(output, to, time);
}

void commands(S32& time, Header type, Input& input, ObjectPool& pool, Config& config)
{
	Id sender;
	while(sender = input.get(), sender != 0) {
		S32 size = input.get();
		Input::Cursor start = input.cursor();

		Object* o = pool.get(sender);	//[hfAsender  ID ɎIuWFNgoB
		MovingObject* agent = dynamic_cast<MovingObject*>(o);
		if(agent != 0) {
			switch(type) {
			case AK_LOAD:
				load(time, agent, input, pool);
				break;
			case AK_UNLOAD:
				unload(time, agent, input, pool);
				break;
			case AK_RESCUE:
				rescue(time, agent, input, pool);
				break;
			case AK_STRETCH:
				stretch(time, agent, input, pool, config);
				break;
			case AK_MOVE:
				move(time, agent, input, pool, config);
				break;
			case AK_CLEAR:
				clear(time, agent, input, pool, config);
				break;
			}
		}
		
		input.setCursor(start);
		input.skip(size);
	}
}

void load(S32& time, MovingObject* agent, Input& input, ObjectPool& pool)
{
	if(dynamic_cast<AmbulanceTeam*>(agent) != 0) {
		if(agent->contents().empty()) {
			Id id = input.get();
			MovingObject* target = dynamic_cast<MovingObject*>(pool.get(id));
			if(target != 0) {
				if(target->position() == agent->position()) {
					target->setPosition(time, agent);
					target->setPositionExtra(time, 0);
					printf("load %ld by %ld\n", (long)target->id(), (long)agent->id());
				}
			}
		}
	}
}
void unload(S32& time, MovingObject* agent, Input& input, ObjectPool& pool)
{
	if(dynamic_cast<AmbulanceTeam*>(agent) != 0) {
		if(!agent->contents().empty()) {
			Object* target = agent->contents().front();
			ASSERT(dynamic_cast<MovingObject*>(target) != 0);
			MovingObject* mo = (MovingObject*)target;
			mo->setPosition(time, agent->position());
			mo->setPositionExtra(time, agent->positionExtra());
			printf("unload %ld by %ld\n", (long)mo->id(), (long)agent->id());
		}
	}
}
void rescue(S32& time, MovingObject* agent, Input& input, ObjectPool& pool)
{
	if(dynamic_cast<AmbulanceTeam*>(agent) != 0) {
		Id id = input.get();
		Humanoid* target = dynamic_cast<Humanoid*>(pool.get(id));
		if(target != 0) {
			if(target->position() == agent->position()) {
				target->setBuriedness(time, target->buriedness() - 1);
				printf("rescue %ld by %ld\n", (long)target->id(), (long)agent->id());
			}
		}
	}
}
void stretch(S32& time, MovingObject* agent, Input& input, ObjectPool& pool, Config& config)
{
	FireBrigade* brigade = dynamic_cast<FireBrigade*>(agent);
	if(brigade != 0) {
		brigade->setStretchedLength(time, brigade->stretchedLength() + config.stretch_length());
		printf("stretch by %ld\n", (long)agent->id());
	}
}
void move(S32& time, MovingObject* agent, Input& input, ObjectPool& pool, Config& config)
{
	FireBrigade* brigade = dynamic_cast<FireBrigade*>(agent);
	if(brigade != 0 && brigade->stretchedLength() > 1)
		brigade->setStretchedLength(time, 0);
}
void clear(S32& time, MovingObject* agent, Input& input, ObjectPool& pool, Config& config)
{
	if(dynamic_cast<PoliceForce*>(agent) != 0) {
		Id id = input.get();
		Road* target = dynamic_cast<Road*>(pool.get(id));
		if(target != 0) {
			if(target != agent->position()	&& target->head() != agent->position() && target->tail() != agent->position())
				return;
			printf("clear %ld by %ld\n", (long)target->id(), (long)agent->id());
			S32 block = target->block();
			S32 repairCost = target->repairCost();
			if(repairCost <= 0) {
				block = 0;
			} else {
				block = block * (repairCost - 1) / repairCost;
				repairCost--;
				target->setRepairCost(time, repairCost);
			}
			target->setBlock(time, block);
			printf("clear %ld by %ld\n", (long)target->id(), (long)agent->id());
		}
	}
}

void step(S32& time, ObjectPool& pool, Config& config)
{
#if 0	// ̃[Xɂ͊܂܂ȂƂɂ
	Objects::const_iterator it = pool.objects().begin();
	for(; it != pool.objects().end(); it++) {
		// PK̐is
		Humanoid* h = dynamic_cast<Humanoid*>(*it);
		if(h != 0) {
			S32 newHp = h->hp() - h->damage();
			if(newHp < 0)
				newHp = 0;
			if(newHp != h->hp())
				h->setHp(time, newHp);
		}
		
		Building* b = dynamic_cast<Building*>(*it);
		if(b != 0) {
			// |ɂ镉
			if(b->brokenness() > b->oldBrokenness()) {
				// |󂵂A
				Objects::const_iterator it2 = b->contents().begin();
				for(; it2 != b->contents().end(); it2++) {
					Humanoid* h = dynamic_cast<Humanoid*>(*it);
					if(h != 0)
						h->setDamage(time, 100);	// G[WFgփ_[W^B
				}
			}
			// ΍Ђɂ镉
			if(1 <= b->fieryness() && b->fieryness() <= 3) {
				// RĂA
				Objects::const_iterator it2 = b->contents().begin();
				for(; it2 != b->contents().end(); it2++) {
					Humanoid* h = dynamic_cast<Humanoid*>(*it);
					if(h != 0)
						h->setDamage(time, 100);	// G[WFgփ_[W^B
				}
			}
		}
	}
#endif
}
