
#include "common.hxx"
#include "basic.hxx"
#include "Config.hxx"
#include "Address.hxx"
#include "Input.hxx"
#include "Output.hxx"
#include "LongUDPSocket.hxx"
#include "RescueObjectPool.hxx"
#include "objects.hxx"
#include "FlammableBuilding.hxx"
using namespace Rescue;

int main2(int argc, char** argv);
void commands(S32& time, Input& input, Output& output, LongUDPSocket& socket, Address& to, RescueObjectPool& pool, Config& config);
void commands(S32& time, Header type, Input& input, RescueObjectPool& pool, Config& config);
void extinguish(S32& time, MovingObject* agent, Input& input, ObjectPool& pool, Config& config);
void step(S32& time, ObjectPool& pool, Config& config);

int main(int argc, char** argv)
{
	try {
		return main2(argc, argv);
	} catch(std::exception& e) {
		fprintf(stderr, "%s\n", (const char*)e.what());
	} catch(...) {
		fprintf(stderr, "error...\n");
	}
	return 1;
}

int main2(int argc, char** argv)
{
	const char* host = "localhost";
	Config config;

	// 
	int i;
	int count = 0;
	for(i=1; i<argc; i++) {
		if(argv[i][0] == '-') {
			if(strcmp(argv[i], "-file") == 0) {
				if(++i < argc) {
					config.readConfigFile(argv[i]);
				} else {
					fprintf(stderr, "error: '-file'\n");
				}
			} else {
				if(++i < argc) {
					if(!config.setValue(argv[i-1]+1, argv[i]))
						fprintf(stderr, "error: '%s'\n", argv[i]);
				} else {
					fprintf(stderr, "error: '%s'\n", argv[i-1]);
				}
			}
		} else {
			switch(count++) {
			case 0:
				host = argv[i];
				break;
			default:
				fprintf(stderr, "error: '%s'\n", argv[i]);
				break;
			}
		}
	}

	// \Pbg쐬
	LongUDPSocket socket(config.textlogname_samplesimulator());
	Input input;
	Output output;
	output.setWait(config.send_udp_wait());
	Address to(host, config.port());	//M

	// ڑbZ[W𑗐M
	printf("connecting...\n");
	output.clear();								//o̓obt@ɂ
	output.put(SK_CONNECT);						//o̓obt@ SK_CONNECT (4oCg)o͂
	Output::Cursor base = output.put(~(S32)0);	//body ̃TCYƂĂƂ肠_~[̒lo͂ĂāAbase ɂ̈ʒuLB
	output.put(0);								//o̓obt@ 0 o͂
	output.setSize(base);						//base ŋLʒuɁAbase 炱܂ł put oCgݒ肷B
	output.put(HEADER_NULL);
	socket.send(output, to, -1);					//output  put oCg to ֑MB

	// bZ[W[v
	S32 time = 0;
	RescueObjectPool pool;
	for(;;) {
		try {
			socket.receive(input);	// TODO: G[OŕԂ	//f[^MAinput ֒~B
			for(;;) {
				Header header = input.get();	//input (4oCg)f[^oAheader ֑
				if(header == HEADER_NULL)
					break;

				S32 size = input.get();
				Input::Cursor start = input.cursor();	//ݓǂݏoĂʒuL

				switch(header) {
				default:
					printf("Unknown header.\n");
					break;
				case KS_CONNECT_ERROR:
					printf("Connection failed.\n");
					return 1;
				case KS_CONNECT_OK:
					printf("Connect Ok.\n");
					// ack Ԃ
					output.clear();
					output.put(SK_ACKNOWLEDGE);
					output.put(~(S32)0);
					output.put(HEADER_NULL);
					socket.send(output, to, -1);

					printf("Initializing...\n");
					to = socket.addressRecievedFrom();	//ȍ~̑MύX
					pool.restructure(0, input);			//[hfB
					pool.setUpMesh(config.mesh_size(), config.mesh_size());
					printf("Initialized.\n");
					break;
				case KS_COMMANDS:
					commands(time, input, output, socket, to, pool, config);
					printf("time = %ld.\n", (long)time);
					break;
				case KS_UPDATE:
					time = input.get();
					pool.restructure(time, input);
					break;
				}

				input.setCursor(start);
				input.skip(size);						//size oCgXLbv
			}
		} catch(std::exception& e) {
			fprintf(stderr, "%s\n", (const char*)e.what());
		} catch(...) {
			fprintf(stderr, "error...\n");
		}
	}
	
	return 0;
}

void commands(S32& time, Input& input, Output& output, LongUDPSocket& socket, Address& to, RescueObjectPool& pool, Config& config)
{
	// KS_COMMANDS M
	time = input.get();
	Header type;
	while(type = input.get(), type != HEADER_NULL) {
		S32 size = input.get();
		Input::Cursor start = input.cursor();
		commands(time, type, input, pool, config);
		input.setCursor(start);
		input.skip(size);
	}

	// Xebv
	step(time, pool, config);

	// V~[Vʂ𑗐M
	output.clear();
	output.put(SK_UPDATE);
	Output::Cursor base = output.put(~(S32)0);
	pool.output(output, time);
	output.setSize(base);
	output.put(HEADER_NULL);
	socket.send(output, to, time);
}

void commands(S32& time, Header type, Input& input, RescueObjectPool& pool, Config& config)
{
	Id sender;
	while(sender = input.get(), sender != 0) {
		S32 size = input.get();
		Input::Cursor start = input.cursor();

		Object* o = pool.get(sender);	//[hfAsender  ID ɎIuWFNgoB
		MovingObject* agent = dynamic_cast<MovingObject*>(o);
		if(agent != 0) {
			switch(type) {
			case AK_EXTINGUISH:
				extinguish(time, agent, input, pool, config);
				break;
			}
		}
		
		input.setCursor(start);
		input.skip(size);
	}
}

void extinguish(S32& time, MovingObject* agent, Input& input, ObjectPool& pool, Config& config)
{
	Id targetID = input.get();
	Object* target = pool.get(targetID);
	FlammableBuilding* fb = dynamic_cast<FlammableBuilding*>(target);
	if(fb != 0)
		fb->extinguish();
}

void step(S32& time, ObjectPool& pool, Config& config)
{
	Objects::const_iterator it = pool.objects().begin();
	for(; it != pool.objects().end(); it++) {
		FlammableBuilding* fb = dynamic_cast<FlammableBuilding*>(*it);
		if(fb != 0) {
			if(fb->ignition() != 0) {
				fb->ignite();
				fb->setIgnition(time, 0);
			}
			fb->step(time);
		}
	}
}
