
#include "common.hxx"
#include "basic.hxx"
#include "Config.hxx"
#include "Address.hxx"
#include "Input.hxx"
#include "Output.hxx"
#include "LongUDPSocket.hxx"
#include "ObjectPool.hxx"
#include "objects.hxx"
using namespace Rescue;

int main2(int argc, char** argv);
void makeWorld(ObjectPool& pool, int number);

int main(int argc, char** argv)
{
	try {
		return main2(argc, argv);
	} catch(std::exception& e) {
		fprintf(stderr, "%s\n", (const char*)e.what());
	} catch(...) {
		fprintf(stderr, "error...\n");
	}
	return 1;
}

int main2(int argc, char** argv)
{
	const char* host = "localhost";
	Config config;
	int number = 50;

	// 
	int i;
	for(i=1; i<argc; i++) {
		if(argv[i][0] == '-') {
			if(strcmp(argv[i], "-file") == 0) {
				if(++i < argc) {
					config.readConfigFile(argv[i]);
				} else {
					fprintf(stderr, "error: '-file'\n");
				}
			} else if(strcmp(argv[i], "-n") == 0) {
				if(++i < argc) {
					number = atoi(argv[i]);
				} else {
					fprintf(stderr, "error: '-file'\n");
				}
			} else {
				if(++i < argc) {
					if(!config.setValue(argv[i-1]+1, argv[i]))
						fprintf(stderr, "error: '%s'\n", argv[i]);
				} else {
					fprintf(stderr, "error: '%s'\n", argv[i-1]);
				}
			}
		} else {
			fprintf(stderr, "error: '%s'\n", argv[i]);
		}
	}

	// \Pbg쐬
	Address address(INADDR_ANY, config.gis_port(), true);
	LongUDPSocket socket(address, config.textlogname_samplegis());
	Input input;
	Output output;
	output.setWait(config.send_udp_wait());
	Address to(host, config.port());	//M

	printf("start\n");

	// bZ[W[v
	S32 time = 0;
	ObjectPool pool;
	makeWorld(pool, number);
	for(;;) {
		try {
			socket.receive(input);	// TODO: G[OŕԂ	//f[^MAinput ֒~B
			for(;;) {
				Header header = input.get();	//input (4oCg)f[^oAheader ֑
				if(header == HEADER_NULL)
					break;

				S32 size = input.get();
				Input::Cursor start = input.cursor();	//ݓǂݏoĂʒuL

				switch(header) {
				default:
					printf("Unknown header. %lx\n", (long)header);
					break;
				case KG_CONNECT:
					printf("KG_CONNECT\n");
					to = socket.addressRecievedFrom();
					output.clear();
					output.put(GK_CONNECT_OK);
					output.put(~(S32)0);
					pool.output(output, 0);
					output.put(HEADER_NULL);
					socket.send(output, to);
					break;
				case KG_ACKNOWLEDGE:
					printf("KG_ACKNOWLEDGE\n");
					break;
				case KG_UPDATE:
					printf("KG_UPDATE\n");
					time = input.get();
					pool.restructure(time, input);
					break;
				}

				input.setCursor(start);
				input.skip(size);						//size oCgXLbv
			}
		} catch(std::exception& e) {
			fprintf(stderr, "%s\n", (const char*)e.what());
		} catch(...) {
			fprintf(stderr, "error...\n");
		}
	}
	
	return 0;
}

void makeWorld(ObjectPool& pool, int number)
{
		enum { testSize = 10 };
		Id id = 1;
		Node* testIntersection[testSize][testSize];
		Road* testRoads[testSize * (testSize-1) * 2];

		const int unit = 50 * 1000;	//1500 * 1000 / (testSize + 2);

		int i, j, k;
		for(i=0; i<COUNT_OF(testIntersection); i++) {
			for(j=0; j<COUNT_OF(testIntersection[i]); j++) {
				Node* node = new Node();
				node->setX(0, i * unit + 1 * unit);
				node->setY(0, j * unit + 1 * unit);
				node->setId(id++);
				pool.add(node);
				testIntersection[i][j] = node;
			}
		}
		for(i=0; i<COUNT_OF(testIntersection); i++) {
			for(j=0; j<COUNT_OF(testIntersection[i]); j++) {
				Building* building;
				switch(j % 1) {
				default:
					building = new Building();
					break;
				case 0:
					building = new Refuge();
					break;
				case 1:
					building = new PoliceOffice();
					break;
				case 2:
					building = new AmbulanceCenter();
					break;
				case 3:
					building = new FireStation();
					break;
				}
				
				
				building->setX(0, i * unit + 1 * unit + unit/3);
				building->setY(0, j * unit + 1 * unit + unit/3);
				//building->setFieryness(0, i % 7);
				building->setId(id++);
				pool.add(building);
				testIntersection[i][j]->testConnect(0, building);
			}
		}
		for(i=0; i<COUNT_OF(testRoads); i++) {
			Road* road = new Road();
			road->setId(id++);
			pool.add(testRoads[i] = road);
		}
		
		k=0;
		for(i=0; i<COUNT_OF(testIntersection)-1; i++) {
			for(j=0; j<COUNT_OF(testIntersection[i]); j++) {
				Road* road = testRoads[k++];
				Node* node1 = testIntersection[i][j];
				Node* node2 = testIntersection[i+1][j];
				ASSERT(road && node1 && node2);
				node1->testConnect(0, road, true);
				node2->testConnect(0, road, false);
				double x = (double)node1->x() - node2->x();
				double y = (double)node1->y() - node2->y();
				road->setLength(0, (S32)sqrt(x*x + y*y));
			}
		}
		for(i=0; i<COUNT_OF(testIntersection); i++) {
			for(j=0; j<COUNT_OF(testIntersection[i])-1; j++) {
				Road* road = testRoads[k++];
				Node* node1 = testIntersection[i][j];
				Node* node2 = testIntersection[i][j+1];
				ASSERT(road && node1 && node2);
				node1->testConnect(0, road, true);
				node2->testConnect(0, road, false);
				double x = (double)node1->x() - node2->x();
				double y = (double)node1->y() - node2->y();
				road->setLength(0, (S32)sqrt(x*x + y*y));
			}
		}

		int z;
#if 1
		k = number;
		z = COUNT_OF(testRoads) / k;
		if(z < 1)
			z = 1;
		for(i=0; i<COUNT_OF(testRoads); i++) {
			if(i % z == 0 && k-- > 0) {
				Road* road = testRoads[i];
				Civilian* civilian = new Civilian();
				civilian->setId(id++);
				pool.add(civilian);
				civilian->setPosition(0, road);
				civilian->setPositionExtra(0, road->length() * 3 / 10);
				//civilian->setDamage(0, i * 5);
			}
		}
#endif
#if 1
		k = 2;
		z = COUNT_OF(testIntersection) / k;
		if(z < 1)
			z = 1;
		for(i=1; i<COUNT_OF(testIntersection); i++) {
			if(i % z == 0 && k-- > 0) {
				Node* node = testIntersection[i][0];
				AmbulanceTeam* ambulance = new AmbulanceTeam();
				ambulance->setId(id++);
				pool.add(ambulance);
				ambulance->setPosition(0, node);
			}
		}
#endif
}
