
#include "common.hxx"
#include "basic.hxx"
#include "Config.hxx"
#include "Address.hxx"
#include "Input.hxx"
#include "Output.hxx"
#include "LongUDPSocket.hxx"
#include "ObjectPool.hxx"
#include "objects.hxx"
using namespace Rescue;

int main2(int argc, char** argv);
void commands(S32& time, Input& input, Output& output, LongUDPSocket& socket, Address& to, ObjectPool& pool);
void commands(S32& time, Header type, Input& input, ObjectPool& pool);
void move(S32& time, MovingObject* agent, Input& input, ObjectPool& pool);

int main(int argc, char** argv)
{
	try {
		return main2(argc, argv);
	} catch(std::exception& e) {
		fprintf(stderr, "%s\n", (const char*)e.what());
	} catch(...) {
		fprintf(stderr, "error...\n");
	}
	return 1;
}

int main2(int argc, char** argv)
{
	const char* host = "localhost";
	Config config;

	// 
	int i;
	int count = 0;
	for(i=1; i<argc; i++) {
		if(argv[i][0] == '-') {
			if(strcmp(argv[i], "-file") == 0) {
				if(++i < argc) {
					config.readConfigFile(argv[i]);
				} else {
					fprintf(stderr, "error: '-file'\n");
				}
			} else {
				if(++i < argc) {
					if(!config.setValue(argv[i-1]+1, argv[i]))
						fprintf(stderr, "error: '%s'\n", argv[i]);
				} else {
					fprintf(stderr, "error: '%s'\n", argv[i-1]);
				}
			}
		} else {
			switch(count++) {
			case 0:
				host = argv[i];
				break;
			default:
				fprintf(stderr, "error: '%s'\n", argv[i]);
				break;
			}
		}
	}

	// \Pbg쐬
	LongUDPSocket socket(config.textlogname_samplesimulator());
	Input input;
	Output output;
	output.setWait(config.send_udp_wait());
	Address to(host, config.port());	//M

	// ڑbZ[W𑗐M
	printf("connecting...\n");
	output.clear();								//o̓obt@ɂ
	output.put(SK_CONNECT);						//o̓obt@ SK_CONNECT (4oCg)o͂
	Output::Cursor base = output.put(~(S32)0);	//body ̃TCYƂĂƂ肠_~[̒lo͂ĂāAbase ɂ̈ʒuLB
	output.put(0);								//o̓obt@ 0 o͂
	output.setSize(base);						//base ŋLʒuɁAbase 炱܂ł put oCgݒ肷B
	output.put(HEADER_NULL);
	socket.send(output, to, -1);					//output  put oCg to ֑MB

	// bZ[W[v
	S32 time = 0;
	ObjectPool pool;
	for(;;) {
		try {
			socket.receive(input);	// TODO: G[OŕԂ	//f[^MAinput ֒~B
			for(;;) {
				Header header = input.get();	//input (4oCg)f[^oAheader ֑
				if(header == HEADER_NULL)
					break;

				S32 size = input.get();
				Input::Cursor start = input.cursor();	//ݓǂݏoĂʒuL

				switch(header) {
				default:
					printf("Unknown header.\n");
					break;
				case KS_CONNECT_ERROR:
					printf("Connection failed.\n");
					return 1;
				case KS_CONNECT_OK:
					printf("Connect Ok.\n");
					// ack Ԃ
					output.clear();
					output.put(SK_ACKNOWLEDGE);
					output.put(~(S32)0);
					output.put(HEADER_NULL);
					socket.send(output, to, -1);

					to = socket.addressRecievedFrom();	//ȍ~̑MύX
					pool.restructure(0, input);			//[hfB
					break;
				case KS_COMMANDS:
					commands(time, input, output, socket, to, pool);
					printf("time = %ld.\n", (long)time);
					break;
				case KS_UPDATE:
					time = input.get();
					pool.restructure(time, input);
					break;
				}

				input.setCursor(start);
				input.skip(size);						//size oCgXLbv
			}
		} catch(std::exception& e) {
			fprintf(stderr, "%s\n", (const char*)e.what());
		} catch(...) {
			fprintf(stderr, "error...\n");
		}
	}
	
	return 0;
}

void commands(S32& time, Input& input, Output& output, LongUDPSocket& socket, Address& to, ObjectPool& pool)
{
	// KS_COMMANDS M
	time = input.get();
	Header type;
	while(type = input.get(), type != HEADER_NULL) {
		S32 size = input.get();
		Input::Cursor start = input.cursor();
		commands(time, type, input, pool);
		input.setCursor(start);
		input.skip(size);
	}

	// V~[Vʂ𑗐M
	output.clear();
	output.put(SK_UPDATE);
	Output::Cursor base = output.put(~(S32)0);
	pool.output(output, time);
	output.setSize(base);
	output.put(HEADER_NULL);
	socket.send(output, to, time);
}

void commands(S32& time, Header type, Input& input, ObjectPool& pool)
{
	Id sender;
	while(sender = input.get(), sender != 0) {
		S32 size = input.get();
		Input::Cursor start = input.cursor();

		Object* o = pool.get(sender);	//[hfAsender  ID ɎIuWFNgoB
		MovingObject* agent = dynamic_cast<MovingObject*>(o);
		if(agent != 0) {
			switch(type) {
			case AK_MOVE:
				move(time, agent, input, pool);
				break;
			}
		}
		
		input.setCursor(start);
		input.skip(size);
	}
}

void move(S32& time, MovingObject* agent, Input& input, ObjectPool& pool)
{
	// agent ̈ړV~[Vs
	MotionlessObject* position = dynamic_cast<MotionlessObject*>(agent->position());	// ݈ʒu
	if(position != 0) {
		S32 extra = agent->positionExtra();		// ݘHɂꍇAhead ̋AȂ 0
		double rest = 60.0;	// Ƃ̕bV~[VȂ΂ȂȂ
		Id id;
		while(id = input.get(), id != 0) {
			ASSERT(dynamic_cast<MotionlessObject*>(pool.get(id)) != 0);
			MotionlessObject* next = (MotionlessObject*)pool.get(id);
			while(next != position) {
				MotionlessObject* nose = next;
				Objects::const_iterator first = position->neighbors().begin();
				Objects::const_iterator last = position->neighbors().end();
				if(std::find(first, last, nose) == last) {
					Objects::const_iterator it = first;
					for(; it != last; it++) {
						MotionlessObject* mo = (MotionlessObject*)*it;
						Objects::const_iterator first = mo->neighbors().begin();
						Objects::const_iterator last = mo->neighbors().end();
						if(std::find(first, last, nose) != last)
							break;
					}
					if(it == last)
						goto label;	// ȍs
					nose = (MotionlessObject*)*it;
				}
				
				Road* road = dynamic_cast<Road*>(position);
				if(road != 0) {
					// Hɂ
					S32 length = road->length();
					if(road->tail() == nose)
						extra = length - extra;		//  Node ܂ł̋
					double mobility = rest * 0.5 * 1000;	// b0.5*1000mmŁAƉmړł邩B
					if(road->contents().size() > 1)
						mobility /= road->contents().size();	// ɂlɔႷ邱Ƃɂ
					mobility += 1;	// ǂȂɂ܂ĂAƂ͓邱Ƃɂ
					if(mobility < extra) {
						// ςOɃ^CAbv
						extra -= (S32)mobility;
						rest = 0;
					} else {
						// ^CAbv̑OɁÃIuWFNgֈړ
						rest -= rest * extra / mobility;
						extra = 0;
					}
					if(road->tail() == nose)
						extra = length - extra;
					if(rest <= 0)
						break;		// ^CAbv
				}
				
				road = dynamic_cast<Road*>(nose);
				if(road != 0) {
					// ̏uԘHɂ
					if(road->head() == position)
						extra = 0;
					else
						extra = road->length();
				} else {
					extra = 0;
				}
				position = nose;
			}
		}
label:
		agent->setPosition(time, position);
		agent->setPositionExtra(time, extra);
	}
}
