
#include "common.hxx"
#include "basic.hxx"
#include "Config.hxx"
#include "Address.hxx"
#include "Input.hxx"
#include "Output.hxx"
#include "LongUDPSocket.hxx"
#include "ObjectPool.hxx"
#include "objects.hxx"
#include "sim_class.h"
using namespace Rescue;

int main2(int argc, char** argv);
void initRealMap(ObjectPool& pool, RealMap& real_map);
void commands(S32& time, Input& input, Output& output, LongUDPSocket& socket, Address& to, ObjectPool& pool, RealMap& real_map);
void commands(S32& time, Header type, Input& input, ObjectPool& pool, RealMap& real_map);
void setMotionOnRoad(MovingObject* agent, Input& input, ObjectPool& pool, RealMap& real_map);
void setMotionOnNode(MovingObject* agent, Input& input, ObjectPool& pool, RealMap& real_map);
void setStay(MovingObject* agent, RealMap& real_map);
//void move(S32& time, MovingObject* agent, Input& input, ObjectPool& pool);
void reviseRealMap(RealMap& real_map);

int main(int argc, char** argv)
{
	try {
		return main2(argc, argv);
	} catch(std::exception& e) {
		fprintf(stderr, "%s\n", (const char*)e.what());
	} catch(...) {
		fprintf(stderr, "error...\n");
	}
	return 1;
}

int main2(int argc, char** argv)
{
	const char* host = "localhost";
	Config config;

	// 
	int i;
	int count = 0;
	for(i=1; i<argc; i++) {
		if(argv[i][0] == '-') {
			if(strcmp(argv[i], "-file") == 0) {
				if(++i < argc) {
					config.readConfigFile(argv[i]);
				} else {
					fprintf(stderr, "error: '-file'\n");
				}
			} else {
				if(++i < argc) {
					if(!config.setValue(argv[i-1]+1, argv[i]))
						fprintf(stderr, "error: '%s'\n", argv[i]);
				} else {
					fprintf(stderr, "error: '%s'\n", argv[i-1]);
				}
			}
		} else {
			switch(count++) {
			case 0:
				host = argv[i];
				break;
			default:
				fprintf(stderr, "error: '%s'\n", argv[i]);
				break;
			}
		}
	}

	// \Pbg쐬
	LongUDPSocket socket(config.textlogname_samplesimulator());
	Input input;
	Output output;
	output.setWait(config.send_udp_wait());
	Address to(host, config.port());	//M

	// ڑbZ[W𑗐M
	printf("connecting...\n");
	output.clear();								//o̓obt@ɂ
	output.put(SK_CONNECT);						//o̓obt@ SK_CONNECT (4oCg)o͂
	Output::Cursor base = output.put(~(S32)0);	//body ̃TCYƂĂƂ肠_~[̒lo͂ĂāAbase ɂ̈ʒuLB
	output.put(0);								//o̓obt@ 0 o͂
	output.setSize(base);						//base ŋLʒuɁAbase 炱܂ł put oCgݒ肷B
	output.put(HEADER_NULL);
	socket.send(output, to, -1);					//output  put oCg to ֑MB

	// bZ[W[v
	S32 time = 0;
	ObjectPool pool;

	// V~[^̃IuWFNg錾
	RealMap real_map;

	for(;;) {
		try {
			socket.receive(input);	// TODO: G[OŕԂ	//f[^MAinput ֒~B
			for(;;) {
				Header header = input.get();	//input (4oCg)f[^oAheader ֑
				if(header == HEADER_NULL)
					break;

				S32 size = input.get();
				Input::Cursor start = input.cursor();	//ݓǂݏoĂʒuL

				switch(header) {
				default:
					printf("Unknown header.\n");
					break;
				case KS_CONNECT_ERROR:
					printf("Connection failed.\n");
					return 1;
				case KS_CONNECT_OK:
					printf("Connect Ok.\n");
					// ack Ԃ
					output.clear();
					output.put(SK_ACKNOWLEDGE);
					output.put(~(S32)0);
					output.put(HEADER_NULL);
					socket.send(output, to, -1);

					to = socket.addressRecievedFrom();	//ȍ~̑MύX
					pool.restructure(0, input);			//[hfB
					initRealMap(pool, real_map);
					break;
				case KS_COMMANDS:
					commands(time, input, output, socket, to, pool, real_map);
					printf("time = %ld.\n", (long)time);
					break;
				case KS_UPDATE:
					time = input.get();
					pool.restructure(time, input);
					break;
				}

				input.setCursor(start);
				input.skip(size);						//size oCgXLbv
			}
		} catch(std::exception& e) {
			fprintf(stderr, "%s\n", (const char*)e.what());
		} catch(...) {
			fprintf(stderr, "error...\n");
		}
	}
	
	return 0;
}

//  V~[^IuWFNg̏ 
void initRealMap(ObjectPool& pool, RealMap& real_map)
{
  Objects::const_iterator p;
  SimCars::iterator p_scar;
  SimNodes::iterator p_snode;
  SimRoads::iterator p_sroad;

  //printf("000\n");
  // pool̃|C^IuWFNgɊUăRs[
  for(p = pool.objects().begin(); p != pool.objects().end(); p++){

    // G[WFg
    MovingObject* p_mo = dynamic_cast<MovingObject*>(*p);
    if(p_mo != 0){
      real_map.sim_car.insert(make_pair(p_mo->id(),SimCar(p_mo)));
      continue;
    }
    // m[h
    Node* p_node = dynamic_cast<Node*>(*p);
    if(p_node != 0){
      real_map.sim_node.insert(make_pair(p_node->id(),SimNode(p_node)));
      continue;
    }
    // [h
    Road* p_road = dynamic_cast<Road*>(*p);
    if(p_road != 0){
      real_map.sim_road.insert(make_pair(p_road->id(),SimRoad(p_road)));
      continue;
    }
  }

  // ԍŏ̏
  for(p_snode = real_map.sim_node.begin(); p_snode != real_map.sim_node.end(); p_snode++){
    p_snode->second.setEdges();
    p_snode->second.setAccessKind();
    p_snode->second.setShortcutToTurn();
    p_snode->second.setPocketToTurnAcross();
    p_snode->second.setSignalTiming();
  }
  for(p_sroad = real_map.sim_road.begin(); p_sroad != real_map.sim_road.end(); p_sroad++){
    p_sroad->second.setHead();
    p_sroad->second.setTail();
    p_sroad->second.setHeadConnectNum(real_map.sim_node);
    p_sroad->second.setTailConnectNum(real_map.sim_node);
  }
}

void commands(S32& time, Input& input, Output& output, LongUDPSocket& socket, Address& to, ObjectPool& pool, RealMap& real_map)
{
  // 芸S̃G[WFgpositionpositionExtraRs[
  SimCars::iterator p;
  for(p = real_map.sim_car.begin(); p != real_map.sim_car.end(); p++){
    p->second.setPosition();
    p->second.setPositionExtra();
  }

	// KS_COMMANDS M
	time = input.get();
	Header type;
	while(type = input.get(), type != HEADER_NULL) {
		S32 size = input.get();
		Input::Cursor start = input.cursor();
		commands(time, type, input, pool, real_map);
		input.setCursor(start);
		input.skip(size);
	}

	//  V~[^̐Ȅ 
	reviseRealMap(real_map);

	//  1Ԃ̌ʗV~[V 
	real_map.truffic();

	for(p = real_map.sim_car.begin(); p != real_map.sim_car.end(); p++){
	  if(p->second.motion != 0){
	    p->second.revisePosition(time,pool);
	    p->second.revisePositionExtra(time);
	  }
	}

	// V~[Vʂ𑗐M
	output.clear();
	output.put(SK_UPDATE);
	Output::Cursor base = output.put(~(S32)0);
	pool.output(output, time);
	output.setSize(base);
	output.put(HEADER_NULL);
	socket.send(output, to, time);
}

void commands(S32& time, Header type, Input& input, ObjectPool& pool, RealMap& real_map)
{
	Id sender;
	while(sender = input.get(), sender != 0) {
		S32 size = input.get();
		Input::Cursor start = input.cursor();

		Object* o = pool.get(sender);	//[hfAsender  ID ɎIuWFNgoB
		MovingObject* agent = dynamic_cast<MovingObject*>(o);
		Road* p_road;
		Node* p_node;
		if(agent != 0) {
			switch(type) {
			case AK_MOVE:
			        o = pool.get(agent->position()->id());
			        p_road = dynamic_cast<Road*>(o);
				if(p_road != 0){
				  setMotionOnRoad(agent, input, pool, real_map);
				  break;
				}
				p_node = dynamic_cast<Node*>(o);
				if(p_node != 0){
				  setMotionOnNode(agent, input, pool, real_map);
				  break;
				}
				else setStay(agent,real_map);
			        //move(time, agent, input, pool);
				break;
			default:
				setStay(agent, real_map);
			}
		}
		
		input.setCursor(start);
		input.skip(size);
	}
}

//  moveR}h𑗐MG[WFg̖ړIñXgݒ(ݒnRoad̏ꍇ) 
void setMotionOnRoad(MovingObject* agent, Input& input, ObjectPool& pool, RealMap& real_map)
{
  SimCars::iterator p = real_map.sim_car.find(agent->id());
  //p->second.setPosition();
  //p->second.setPositionExtra();
  p->second.motion = 1;                                         // ړtOON
  p->second.aim.clear();                                        // ړInNA
  p->second.aim.push_back(p->second.position);                  // ݈ʒuaim[0]ɐݒ

  Id aim_id;
  while(aim_id = input.get(), aim_id != 0){
    Object* o = pool.get(aim_id);
    Road* p_road = dynamic_cast<Road*>(o);
    if(p_road != 0){// MꂽړIñXg[ĥ݂L
      p->second.aim.push_back(aim_id);
    }
    //Node* p_node = dynamic_cast<Node*>(o);
    //if(p_node != 0){// MꂽړIñXĝm[h̍ŌL
    //  p->second.last_node = aim_id;
    //}
  }

  p->second.aim_num = p->second.aim.size();                     // ړIn̐
}

//  moveR}h𑗐MG[WFg̖ړIñXgݒ(ݒnNodȅꍇ) 
void setMotionOnNode(MovingObject* agent, Input& input, ObjectPool& pool, RealMap& real_map)
{
  SimCars::iterator p = real_map.sim_car.find(agent->id());
  p->second.motion = 1;                                         // ړtOON
  p->second.aim.clear();                                        // ړInNA  

  Id aim_id;
  aim_id = input.get();
  Object* o = pool.get(aim_id);
  Road* p_road = dynamic_cast<Road*>(o);
  if(p_road != 0){
    p->second.position = aim_id;
    //p->second.aim.push_back(aim_id);
    if((p_road->head()->id()) == (agent->position()->id())){
      p->second.positionExtra = 0;
    }
    else if((p_road->tail()->id()) == (agent->position()->id())){
      p->second.positionExtra = p_road->length();
    }
    else{
      printf("Connection contradiction.\n");
      exit(1);
    }

    while(aim_id = input.get(), aim_id != 0){
      Object* o = pool.get(aim_id);
      Road* p_road = dynamic_cast<Road*>(o);
      if(p_road != 0){// MꂽړIñXg[ĥ݂L
	p->second.aim.push_back(aim_id);
      }
      //Node* p_node = dynamic_cast<Node*>(o);
      //if(p_node != 0){// MꂽړIñXĝm[h̍ŌL
      //  p->second.last_node = aim_id;
      //}
    }

    p->second.aim_num = p->second.aim.size();                     // ړIn̐
  }
  else{
    printf("Agent is on the node, but...\n");
    exit(1);
  }
}

//  ȂG[WFg͈ړtOmotionOFFɂ 
void setStay(MovingObject* agent, RealMap& real_map)
{
  SimCars::iterator p = real_map.sim_car.find(agent->id());
  p->second.motion = 0;
}

#if 0
void move(S32& time, MovingObject* agent, Input& input, ObjectPool& pool)
{
	// agent ̈ړV~[Vs
	MotionlessObject* position = dynamic_cast<MotionlessObject*>(agent->position());	// ݈ʒu
	if(position != 0) {
		S32 extra = agent->positionExtra();		// ݘHɂꍇAhead ̋AȂ 0
		double rest = 60.0;	// Ƃ̕bV~[VȂ΂ȂȂ
		Id id;
		while(id = input.get(), id != 0) {
			ASSERT(dynamic_cast<MotionlessObject*>(pool.get(id)) != 0);
			MotionlessObject* next = (MotionlessObject*)pool.get(id);
			while(next != position) {
				MotionlessObject* nose = next;
				Objects::const_iterator first = position->neighbors().begin();
				Objects::const_iterator last = position->neighbors().end();
				if(std::find(first, last, nose) == last) {
					Objects::const_iterator it = first;
					for(; it != last; it++) {
						MotionlessObject* mo = (MotionlessObject*)*it;
						Objects::const_iterator first = mo->neighbors().begin();
						Objects::const_iterator last = mo->neighbors().end();
						if(std::find(first, last, nose) != last)
							break;
					}
					if(it == last)
						goto label;	// ȍs
					nose = (MotionlessObject*)*it;
				}
				
				Road* road = dynamic_cast<Road*>(position);
				if(road != 0) {
					// Hɂ
					S32 length = road->length();
					if(road->tail() == nose)
						extra = length - extra;		//  Node ܂ł̋
					double mobility = rest * 0.5 * 1000;	// b0.5*1000mmŁAƉmړł邩B
					if(road->contents().size() > 1)
						mobility /= road->contents().size();	// ɂlɔႷ邱Ƃɂ
					mobility += 1;	// ǂȂɂ܂ĂAƂ͓邱Ƃɂ
					if(mobility < extra) {
						// ςOɃ^CAbv
						extra -= (S32)mobility;
						rest = 0;
					} else {
						// ^CAbv̑OɁÃIuWFNgֈړ
						rest -= rest * extra / mobility;
						extra = 0;
					}
					if(road->tail() == nose)
						extra = length - extra;
					if(rest <= 0)
						break;		// ^CAbv
				}
				
				road = dynamic_cast<Road*>(nose);
				if(road != 0) {
					// ̏uԘHɂ
					if(road->head() == position)
						extra = 0;
					else
						extra = road->length();
				} else {
					extra = 0;
				}
				position = nose;
			}
		}
label:
		agent->setPosition(time, position);
		agent->setPositionExtra(time, extra);
	}
}
#endif
