#ifndef SIMCLASS
#define SIMCLASS

namespace Rescue{

struct SimCar{
  MovingObject* pointer;     // 
  Id position;               // 
  S32 positionExtra;         // 
  int flag;                  //
  int direct;                //
  S32 speed;                 //
  S32 between;               //
  Id front_id;               //
  Id back_id;                //
  int step;                  //
  int pocket;                //
  int line;                  //
  std::vector<Id> aim;       // 
  int aim_num;               // 
//  Id last_node;              // 
  int motion;                // 
public:
  SimCar(MovingObject* p_mo){
    pointer = p_mo;
    position = 0;
    positionExtra = 0;
    flag = 0;
    direct = 0;
    speed = 0;
    between = 0;
    front_id = 0;
    back_id = 0;
    step = 0;
    pocket = 0;
    line = 0;
    aim_num = 0;
    motion = 0; 
  }
  void setPosition(){ position = pointer->position()->id(); }
  void setPositionExtra(){ positionExtra = pointer->positionExtra(); }
  void revisePosition(S32& time, ObjectPool& pool){ pointer->setPosition(time,pool.get(position)); }
  void revisePositionExtra(S32& time){ pointer->setPositionExtra(time,positionExtra); }
};

typedef std::map<Id,SimCar> SimCars;

struct SimNode{
  Node* pointer;             // 
  Id edges[6];               // 
  int connect_road_num;      // 
  int accessKind[6][6];      // 
  std::vector<int> shortcutToTurn;
  std::vector<int> pocketToTurnAcross;
  std::vector<int> signalTiming;
public:
  SimNode(Node* p_node){
    pointer = p_node;
    connect_road_num = 0;
  }
  void setEdges(){
    int count = 0;
    Objects::const_iterator p;
    // m_edgesIuWFNĝ[ĥ݂oIDԍedgesɊi[
    for(p = pointer->edges().begin(); p != pointer->edges().end(); p++){
      Road* p_road = dynamic_cast<Road*>(*p);
      if(p_road != 0) edges[count++] = p_road->id();
    }
    connect_road_num = count;
  }
  void setAccessKind(){
    int i;
    switch(connect_road_num){
    case 1:
      accessKind[0][0] = -1;
      break;
    case 2:
      accessKind[0][0] = accessKind[1][1] = -1;
      accessKind[0][1] = accessKind[1][0] =  2;
      break;
    case 3:
      accessKind[0][0] = accessKind[1][1] = accessKind[2][2] = -1;
      accessKind[1][0] = accessKind[2][0] = 0;
      accessKind[0][1] = accessKind[1][2] = 1;
      accessKind[0][2] = accessKind[2][0] = 2;
      break;
    case 4:
      accessKind[0][0] = accessKind[1][1] = accessKind[2][2] = accessKind[3][3] = -1;
      accessKind[0][3] = accessKind[1][0] = accessKind[2][1] = accessKind[3][2] =  0;
      accessKind[0][1] = accessKind[1][2] = accessKind[2][3] = accessKind[3][0] =  1;
      accessKind[0][2] = accessKind[1][3] = accessKind[2][0] = accessKind[3][1] =  2;
      break;
    case 5:
      for(i=0;i<5;i++)  accessKind[i][i] = -1;
      accessKind[0][3] = accessKind[0][4] = accessKind[1][0] = accessKind[1][4] = accessKind[2][1] = accessKind[3][2] = accessKind[4][3] = 0;
      accessKind[0][1] = accessKind[2][1] = accessKind[2][3] = accessKind[2][4] = accessKind[3][0] = accessKind[3][4] = accessKind[4][0] = 1;
      accessKind[0][2] = accessKind[1][3] = accessKind[2][0] = accessKind[3][1] = accessKind[4][1] = accessKind[4][2] = 2;
      break;
    case 6:
      for(i=0;i<6;i++) accessKind[i][i] = -1;
      accessKind[0][4] = accessKind[0][5] = accessKind[1][0] = accessKind[1][5] = accessKind[2][0] = accessKind[2][1] = accessKind[3][1] = accessKind[3][2] = accessKind[4][2] = accessKind[4][3] = accessKind[5][3] = accessKind[5][4] = 0;
      accessKind[0][1] = accessKind[0][2] = accessKind[1][2] = accessKind[1][3] = accessKind[2][3] = accessKind[2][4] = accessKind[3][4] = accessKind[3][5] = accessKind[4][0] = accessKind[4][5] = accessKind[5][0] = accessKind[5][1] = 1;
      accessKind[0][3] = accessKind[1][4] = accessKind[2][5] = accessKind[3][0] = accessKind[4][1] = accessKind[5][2] = 2;
      break;
    default:
      printf("connect_riad_num is too large.\n");
      break;
    }
  }
  void setShortcutToTurn(){
    Bytes::const_iterator p;
    for(p = pointer->shortcutToTurn().begin(); p != pointer->shortcutToTurn().end(); p++){
      shortcutToTurn.push_back((int)*p);
    }
  }
  void setPocketToTurnAcross(){
    Bytes::const_iterator p;
    for(p = pointer->pocketToTurnAcross().begin(); p != pointer->pocketToTurnAcross().end(); p++){
      pocketToTurnAcross.push_back((int)*p);
    }
  }
  void setSignalTiming(){
    Bytes::const_iterator p;
    for(p = pointer->signalTiming().begin(); p != pointer->signalTiming().end(); p++){
      signalTiming.push_back((int)*p);
    }
  }
};

typedef std::map<Id,SimNode> SimNodes;

struct SimRoad{
  Road* pointer;             // 
  Id head;                   // 
  Id tail;                   // 
  S32 linesAliveHead;        // 
  S32 linesAliveTail;        // 
  Id head_pocket;            //
  Id tail_pocket;            //
  Id head_last_id[6];        //
  Id tail_last_id[6];        //
  int head_connect_num;      // 
  int tail_connect_num;      // 
public:
  SimRoad(Road* p_road){
    pointer = p_road;
    head = 0;
    tail = 0;
    linesAliveHead = 0;
    linesAliveTail = 0;
    head_pocket = 0;
    tail_pocket = 0;
    head_connect_num = 0;
    tail_connect_num = 0;
}
  void setHead(){ head = pointer->head()->id(); }
  void setTail(){ tail = pointer->tail()->id(); }
  void setLinesAlive(){
    S32 pass;
    S32 linesAlive;
    double tmp_linesAliveHead;

    switch((char)(pointer->medianStrip())){
    case 0:
      if(pointer->block() != 0){
#if 1
	pass = pointer->width() - pointer->block();
	linesAlive = (S32)(pass/2500);
	tmp_linesAliveHead = linesAlive*(pointer->linesToHead()/(pointer->linesToHead() + pointer->linesToTail()));
	linesAliveHead = ((tmp_linesAliveHead - (S32)tmp_linesAliveHead) < 0.50)
	                   ? (S32)tmp_linesAliveHead : ((S32)tmp_linesAliveHead + 1);
	linesAliveTail = linesAlive - linesAliveHead;
	linesAliveHead = (linesAliveHead < pointer->linesToHead()) ? linesAliveHead : pointer->linesToHead();
	linesAliveTail = (linesAliveTail < pointer->linesToTail()) ? linesAliveTail : pointer->linesToTail();
#else
	linesAliveHead = pointer->linesToHead();
	linesAliveTail = pointer->linesToTail();
#endif
      }
      else{
	linesAliveHead = pointer->linesToHead();
	linesAliveTail = pointer->linesToTail();
      }
      break;
    case 1:
      linesAliveHead = pointer->linesToHead();
      linesAliveTail = pointer->linesToTail();
      break;
    default:
      printf("medianStrip error\n");
      exit(1);
    }
  }
  void setHeadConnectNum(SimNodes& sim_node){
    SimNodes::iterator p = sim_node.find(head);
    int i;
    for(i = 0; i < p->second.connect_road_num; i++){
      if(p->second.edges[i] == pointer->id()){
	head_connect_num = i;
	break;
      }
    }
    if(i == p->second.connect_road_num){
      printf("SimRoad::head_connect_num error\n");
      exit(1);
    }
  }
  void setTailConnectNum(SimNodes& sim_node){
    SimNodes::iterator p = sim_node.find(tail);
    int i;
    for(i = 0; i < p->second.connect_road_num; i++){
      if(p->second.edges[i] == pointer->id()){
	tail_connect_num = i;
	break;
      }
    }
    if(i == p->second.connect_road_num){
      printf("SimRoad::tail_connect_num error\n");
      exit(1);
    }
  }
};

typedef std::map<Id,SimRoad> SimRoads;

struct RealMap{
  SimCars sim_car;
  SimNodes sim_node;
  SimRoads sim_road;
public:
  void InitEnv();
  void InitCar();
  void InitLines();
  void InitFront();
  void InitRoad();
  void truffic();
  void move(Id);
  void move_to_pocket(Id);
  void move_in_pocket(Id);
  void move_normal(Id);
  void move_front(Id);
  void move_at_cross(Id);
  void move_signal(Id);
  void move_on_blue(Id);
  void move_node_across(Id);
  void move_no_signal(Id);
  void move_node(Id);
  void move_inroad(Id);
  int getAccess(Id,Id);
  int judgeAgainist(Id);
  int getAliveLines(Id,Id,Id);
  int getLineInroad(Id);
  int judgeInroad(Id);
  int pass(Id);
};

}
#endif
